import abc
import collections.abc
import functools
import typing as ta


T = ta.TypeVar('T')


##


class AttributePresentError(Exception):
    pass


SetAttrIfPresent: ta.TypeAlias = ta.Literal['overwrite', 'leave', 'raise']


_NO_SET_ATTR_VALUE = object()


@ta.overload
def set_attr(
        obj: ta.Any,
        name: str,
        *,
        if_present: SetAttrIfPresent = 'overwrite',
) -> ta.Callable[[T], T]:
    ...


@ta.overload
def set_attr(
        obj: ta.Any,
        name: str,
        value: T,
        *,
        if_present: SetAttrIfPresent = 'overwrite',
) -> T:
    ...


def set_attr(
        obj,
        name,
        value=_NO_SET_ATTR_VALUE,
        *,
        if_present='overwrite',
):
    if value is _NO_SET_ATTR_VALUE:
        return functools.partial(
            set_attr,
            obj,
            name,
            if_present=if_present,
        )

    if hasattr(obj, name):
        if if_present == 'overwrite':
            pass
        elif if_present == 'leave':
            return value
        elif if_present == 'raise':
            raise AttributePresentError(name)
        else:
            raise ValueError(if_present)

    setattr(obj, name, value)

    return value


##


class AttrOps(abc.ABC):
    class NOT_SET:  # noqa
        def __new__(cls, *args, **kwargs):  # noqa
            raise TypeError

    @abc.abstractmethod
    def getattr(self, obj: ta.Any, name: str, default: ta.Any = NOT_SET) -> ta.Any:
        raise NotImplementedError

    @abc.abstractmethod
    def setattr(self, obj: ta.Any, name: str, value: ta.Any) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def delattr(self, obj: ta.Any, name: str) -> None:
        raise NotImplementedError


#


class StdAttrOps(AttrOps):
    def getattr(self, obj: ta.Any, name: str, default: ta.Any = AttrOps.NOT_SET) -> ta.Any:
        if default is AttrOps.NOT_SET:
            return getattr(obj, name)
        else:
            return getattr(obj, name, default)

    def setattr(self, obj: ta.Any, name: str, value: ta.Any) -> None:
        setattr(obj, name, value)

    def delattr(self, obj: ta.Any, name: str) -> None:
        delattr(obj, name)


STD_ATTR_OPS = StdAttrOps()


#


class DictAttrOps(AttrOps):
    def __init__(self, dct: ta.MutableMapping[str, ta.Any] | None = None) -> None:
        super().__init__()

        if dct is None:
            dct = {}
        self._dct = dct

    def getattr(self, obj: ta.Any, name: str, default: ta.Any = AttrOps.NOT_SET) -> ta.Any:
        try:
            return self._dct[name]
        except KeyError:
            if default is not AttrOps.NOT_SET:
                return default
            raise AttributeError(name) from None

    def setattr(self, obj: ta.Any, name: str, value: ta.Any) -> None:
        self._dct[name] = value

    def delattr(self, obj: ta.Any, name: str) -> None:
        try:
            del self._dct[name]
        except KeyError:
            raise AttributeError(name) from None


##


class TransientDict(collections.abc.MutableMapping):
    def __init__(self) -> None:
        super().__init__()

        self._dct: dict = {}

    def __reduce__(self):
        return (TransientDict, ())

    def __getitem__(self, item):
        return self._dct[item]

    def __setitem__(self, key, value):
        self._dct[key] = value

    def __delitem__(self, key):
        del self._dct[key]

    def __len__(self):
        return len(self._dct)

    def __iter__(self):
        return iter(self._dct)

    def clear(self):
        self._dct.clear()

    def items(self):
        return self._dct.items()

    def keys(self):
        return self._dct.keys()

    def values(self):
        return self._dct.values()

    def __contains__(self, key, /):
        return super().__contains__(key)

    def __eq__(self, other, /):
        raise TypeError(self)

    def __hash__(self):
        raise TypeError(self)


#


_TRANSIENT_DICT_ATTR = '__transient_dict__'


def _get_object_transient_dict(obj: ta.Any) -> TransientDict:
    try:
        return obj.__dict__[_TRANSIENT_DICT_ATTR]
    except KeyError:
        return obj.__dict__.setdefault(_TRANSIENT_DICT_ATTR, TransientDict())


class TransientAttrOps(AttrOps):
    def getattr(self, obj: ta.Any, name: str, default: ta.Any = AttrOps.NOT_SET) -> ta.Any:
        td = _get_object_transient_dict(obj)
        try:
            return td[name]
        except KeyError:
            if default is not AttrOps.NOT_SET:
                return default
            raise AttributeError(name) from None

    def setattr(self, obj: ta.Any, name: str, value: ta.Any) -> None:
        td = _get_object_transient_dict(obj)
        td[name] = value

    def delattr(self, obj: ta.Any, name: str) -> None:
        td = _get_object_transient_dict(obj)
        try:
            del td[name]
        except KeyError:
            raise AttributeError(name) from None


TRANSIENT_ATTR_OPS = TransientAttrOps()

transient_getattr = TRANSIENT_ATTR_OPS.getattr
transient_setattr = TRANSIENT_ATTR_OPS.setattr
transient_delattr = TRANSIENT_ATTR_OPS.delattr



import html
import typing as ta


##


def escape(s: str) -> str:
    return html.escape(s).replace('@', '&#64;')


class Color(ta.NamedTuple):
    r: int
    g: int
    b: int


def gen_rainbow(steps: int) -> list[Color]:
    colors = []
    for r in range(steps):
        colors.append(Color(r * 255 // steps, 255, 0))
    for g in range(steps, 0, -1):
        colors.append(Color(255, g * 255 // steps, 0))
    for b in range(steps):
        colors.append(Color(255, 0, b * 255 // steps))
    for r in range(steps, 0, -1):
        colors.append(Color(r * 255 // steps, 0, 255))
    for g in range(steps):
        colors.append(Color(0, g * 255 // steps, 255))
    for b in range(steps, 0, -1):
        colors.append(Color(0, 255, b * 255 // steps))
    colors.append(Color(0, 255, 0))
    return colors



import sys
import types
import typing as ta

from ... import dataclasses as dc
from ... import lang
from ..origins import HasOrigins
from ..origins import Origin
from ..origins import Origins


HasOriginsT = ta.TypeVar('HasOriginsT', bound=HasOrigins)


##


ORIGIN_NUM_FRAMES = 8
ORIGIN_BASE_OFS = 2

ORIGIN_IGNORED_PACKAGES = frozenset([
    __package__,
    __package__.rpartition('.')[0],

    lang.__name__,
    lang.functions.__name__,

    dc.__name__,
    *[m for m in sys.modules if m.startswith(dc.__name__ + '.')],
])


def is_origin_frame(f: types.FrameType) -> bool:
    gl = f.f_globals
    try:
        pkg = gl['__package__']
    except KeyError:
        pass
    else:
        if pkg in ORIGIN_IGNORED_PACKAGES:
            return False
    return True


def build_origin(ofs: int = 0) -> Origin:
    lst = []  # type: ignore
    cur = sys._getframe(ORIGIN_BASE_OFS + ofs)  # noqa
    while len(lst) < ORIGIN_NUM_FRAMES and cur is not None:
        if is_origin_frame(cur):
            lst.append(cur)
        cur = cur.f_back  # type: ignore
    return Origin(tuple(str(f) for f in lst))


##


ORIGINS_ATTR = '__inject_origins__'


def set_origins(obj: HasOriginsT, origins: Origins) -> HasOriginsT:
    obj.__dict__[ORIGINS_ATTR] = origins
    return obj


class HasOriginsImpl(HasOrigins):
    """Note: inheritors must be dataclasses."""

    @property
    def origins(self) -> Origins:
        return self.__dict__[ORIGINS_ATTR]

    def __post_init__(self) -> None:
        dc.maybe_post_init(super())
        if ORIGINS_ATTR in self.__dict__:
            raise AttributeError('Origin already set')
        set_origins(self, Origins((build_origin(),)))



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import abc
import sys
import typing as ta

from ..lite.maysyncs import make_maysync
from .async_ import AbstractAsyncSubprocesses
from .async_ import _AbstractAsyncSubprocesses
from .run import SubprocessRun
from .run import SubprocessRunOutput
from .sync import AbstractSubprocesses


##


class AbstractMaysyncSubprocesses(_AbstractAsyncSubprocesses, abc.ABC):
    pass


##


class MaysyncSubprocesses(AbstractMaysyncSubprocesses):
    def __init__(
            self,
            subprocesses: AbstractSubprocesses,
            async_subprocesses: AbstractAsyncSubprocesses,
    ) -> None:
        super().__init__()

        self._subprocesses = subprocesses
        self._async_subprocesses = async_subprocesses

    async def run_(self, run: SubprocessRun) -> SubprocessRunOutput:
        return await make_maysync(
            self._subprocesses.run,
            self._async_subprocesses.run,
        )(run).m()

    async def check_call(self, *cmd: str, stdout: ta.Any = sys.stderr, **kwargs: ta.Any) -> None:
        return await make_maysync(
            self._subprocesses.check_call,
            self._async_subprocesses.check_call,
        )(*cmd, **kwargs).m()

    async def check_output(self, *cmd: str, **kwargs: ta.Any) -> bytes:
        return await make_maysync(
            self._subprocesses.check_output,
            self._async_subprocesses.check_output,
        )(*cmd, **kwargs).m()



# ruff: noqa: UP006 UP007 UP045
import abc
import socket
import typing as ta

from ...lite.check import check
from ...sockets.addresses import SocketAddress


##


class FdioHandler(abc.ABC):
    @abc.abstractmethod
    def fd(self) -> int:
        raise NotImplementedError

    #

    @property
    @abc.abstractmethod
    def closed(self) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> None:
        raise NotImplementedError

    #

    def readable(self) -> bool:
        return False

    def writable(self) -> bool:
        return False

    #

    def on_readable(self) -> None:
        raise TypeError

    def on_writable(self) -> None:
        raise TypeError

    def on_error(self, exc: ta.Optional[BaseException] = None) -> None:  # noqa
        pass


class SocketFdioHandler(FdioHandler, abc.ABC):
    def __init__(
            self,
            addr: SocketAddress,
            sock: socket.socket,
    ) -> None:
        super().__init__()

        self._addr = addr
        self._sock: ta.Optional[socket.socket] = sock

    def fd(self) -> int:
        return check.not_none(self._sock).fileno()

    @property
    def closed(self) -> bool:
        return self._sock is None

    def close(self) -> None:
        if self._sock is not None:
            self._sock.close()
        self._sock = None



import inspect

from ..processing.base import ProcessingContext
from ..processing.base import Processor
from ..processing.priority import ProcessorPriority
from ..processing.registry import register_processor_type


##


def _build_cls_doc(cls: type) -> str:
    try:
        text_sig = str(inspect.signature(cls)).replace(' -> None', '')
    except (TypeError, ValueError):
        text_sig = ''
    return cls.__name__ + text_sig


class _LazyClsDocDescriptor:
    def __get__(self, instance, owner):
        if instance is not None:
            owner = instance.__class__
        if not owner:
            raise RuntimeError
        doc = _build_cls_doc(owner)
        owner.__doc__ = doc
        return doc


@register_processor_type(priority=ProcessorPriority.POST_GENERATION)
class DocProcessor(Processor):
    def __init__(
            self,
            ctx: ProcessingContext,
            *,
            lazy: bool = True,
    ) -> None:
        super().__init__(ctx)

        self._lazy = lazy

    def process(self, cls: type) -> type:
        # FIXME: doesn't update doc in subclasses lol, as per stdlib
        if getattr(cls, '__doc__'):
            return cls

        if self._lazy:
            cls.__doc__ = _LazyClsDocDescriptor()  # type: ignore
        else:
            cls.__doc__ = _build_cls_doc(cls)

        return cls



import dataclasses as dc
import typing as ta

from ..generation.base import Generator
from ..generation.base import Plan
from ..generation.base import PlanResult
from ..generation.globals import NONE_GLOBAL
from ..generation.idents import SELF_IDENT
from ..generation.idents import VALUE_IDENT
from ..generation.ops import AddPropertyOp
from ..generation.ops import Op
from ..generation.ops import OpRef
from ..generation.ops import Ref
from ..generation.registry import register_generator_type
from ..generation.utils import SetattrSrcBuilder
from ..processing.base import ProcessingContext
from .fields import InstanceFields


##


@dc.dataclass(frozen=True)
class OverridePlan(Plan):
    @dc.dataclass(frozen=True)
    class Field:
        name: str
        annotation: OpRef[ta.Any]

    fields: tuple[Field, ...]

    frozen: bool


@register_generator_type(OverridePlan)
class OverrideGenerator(Generator[OverridePlan]):
    def plan(self, ctx: ProcessingContext) -> PlanResult[OverridePlan] | None:
        orm = {}

        flds: list[OverridePlan.Field] = []
        for i, f in enumerate(ctx[InstanceFields]):
            if not (f.override or ctx.cs.override):
                continue
            r: OpRef = OpRef(f'override.fields.{i}.annotation')
            orm[r] = f.annotation
            flds.append(OverridePlan.Field(
                f.name,
                r,
            ))

        if not flds:
            return None

        return PlanResult(
            OverridePlan(
                tuple(flds),
                ctx.cs.frozen,
            ),
            orm,
        )

    def generate(self, pl: OverridePlan) -> ta.Iterable[Op]:
        ops: list[Op] = []

        for f in pl.fields:
            op_refs: set[Ref] = {f.annotation}

            get_src = '\n'.join([
                f'def {f.name}({SELF_IDENT}) -> {f.annotation.ident()}:',
                f'    return {SELF_IDENT}.__dict__[{f.name!r}]',
            ])

            set_src: str | None = None
            if not pl.frozen:
                sab = SetattrSrcBuilder()
                set_src = '\n'.join([
                    f'def {f.name}({SELF_IDENT}, {VALUE_IDENT}) -> {NONE_GLOBAL.ident}:',
                    *[
                        f'    {l}'
                        for l in sab(
                            f.name,
                            VALUE_IDENT,
                            frozen=pl.frozen,
                            override=True,
                        )
                    ],
                ])
                op_refs.add(NONE_GLOBAL)
                op_refs.update(sab.refs)

            ops.append(AddPropertyOp(
                f.name,
                get_src=get_src,
                set_src=set_src,
                refs=frozenset(op_refs),
            ))

        return ops



# type: ignore
# ruff: noqa
# flake8: noqa
#
# Copyright (c) 2012-2017 The ANTLR Project. All rights reserved.
# Use of this file is governed by the BSD 3-clause license that
# can be found in the LICENSE.txt file in the project root.
#

#
# A {@link Token} object representing an entire subtree matched by a parser
# rule; e.g., {@code <expr>}. These tokens are created for {@link TagChunk}
# chunks where the tag corresponds to a parser rule.
#
from ..Token import Token


class RuleTagToken(Token):
    __slots__ = ('label', 'ruleName')
    #
    # Constructs a new instance of {@link RuleTagToken} with the specified rule
    # name, bypass token type, and label.
    #
    # @param ruleName The name of the parser rule this rule tag matches.
    # @param bypassTokenType The bypass token type assigned to the parser rule.
    # @param label The label associated with the rule tag, or {@code null} if
    # the rule tag is unlabeled.
    #
    # @exception IllegalArgumentException if {@code ruleName} is {@code null}
    # or empty.

    def __init__(self, ruleName:str, bypassTokenType:int, label:str=None):
        if ruleName is None or len(ruleName)==0:
            raise Exception("ruleName cannot be null or empty.")
        self.source = None
        self.type = bypassTokenType # token type of the token
        self.channel = Token.DEFAULT_CHANNEL # The parser ignores everything not on DEFAULT_CHANNEL
        self.start = -1 # optional; return -1 if not implemented.
        self.stop = -1  # optional; return -1 if not implemented.
        self.tokenIndex = -1 # from 0..n-1 of the token object in the input stream
        self.line = 0 # line=1..n of the 1st character
        self.column = -1 # beginning of the line at which it occurs, 0..n-1
        self.label = label
        self._text = self.getText() # text of the token.

        self.ruleName = ruleName


    def getText(self):
        if self.label is None:
            return "<" + self.ruleName + ">"
        else:
            return "<" + self.label + ":" + self.ruleName + ">"



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import functools
import typing as ta


T = ta.TypeVar('T')


##


def mut_toposort(data: ta.Dict[T, ta.Set[T]]) -> ta.Iterator[ta.Set[T]]:
    for k, v in data.items():
        v.discard(k)
    extra_items_in_deps = functools.reduce(set.union, data.values()) - set(data.keys())
    data.update({item: set() for item in extra_items_in_deps})
    while True:
        ordered = {item for item, dep in data.items() if not dep}
        if not ordered:
            break
        yield ordered
        data = {item: (dep - ordered) for item, dep in data.items() if item not in ordered}
    if data:
        raise ValueError('Cyclic dependencies exist among these items: ' + ' '.join(repr(x) for x in data.items()))


def toposort(data: ta.Mapping[T, ta.AbstractSet[T]]) -> ta.Iterator[ta.Set[T]]:
    return mut_toposort({k: set(v) for k, v in data.items()})
